#usr/bin/python3
# -*- coding: UTF-8 -*-
from . import *
from warnings import warn
import numpy as np
    
def get_para_vector(network) -> torch.Tensor:
    """
    Returns the parameter vector of the given network.

    Args:
        network (torch.nn.Module): The network for which to compute the gradient vector.

    Returns:
        torch.Tensor: The parameter vector of the network.
    """
    with torch.no_grad():
        para_vec = None
        for par in network.parameters():
            viewed=par.data.view(-1)
            if para_vec is None:
                para_vec = viewed
            else:
                para_vec = torch.cat((para_vec, viewed))
        return para_vec
    
def get_gradient_vector(network)->torch.Tensor:
    """
    Returns the gradient vector of the given network.

    Args:
        network (torch.nn.Module): The network for which to compute the gradient vector.

    Returns:
        torch.Tensor: The gradient vector of the network.
    """
    with torch.no_grad():
        grad_vec = None
        for par in network.parameters():
            viewed=par.grad.data.view(-1)
            if grad_vec is None:
                grad_vec = viewed
            else:
                grad_vec = torch.cat((grad_vec, viewed))
        return grad_vec

def apply_gradient_vector(network:torch.nn.Module,grad_vec:torch.Tensor)->None:
    """
    Applies a gradient vector to the network's parameters.

    Args:
        network (torch.nn.Module): The network to apply the gradient vector to.
        grad_vec (torch.Tensor): The gradient vector to apply.

    """
    with torch.no_grad():
        start=0
        for par in network.parameters():
            end=start+par.grad.data.view(-1).shape[0]
            par.grad.data=grad_vec[start:end].view(par.grad.data.shape)
            start=end

def apply_para_vector(network:torch.nn.Module,para_vec:torch.Tensor)->None:
    """
    Applies a parameter vector to the network's parameters.
    
    Args:
        network (torch.nn.Module): The network to apply the parameter vector to.
        para_vec (torch.Tensor): The parameter vector to apply.
    """
    with torch.no_grad():
        start=0
        for par in network.parameters():
            end=start+par.data.view(-1).shape[0]
            par.data=para_vec[start:end].view(par.data.shape)
            start=end

def get_cos_similarity(vector1:torch.Tensor,vector2:torch.Tensor)->torch.Tensor:
    """
    Calculates the cosine angle between two vectors.

    Args:
        vector1 (torch.Tensor): The first vector.
        vector2 (torch.Tensor): The second vector.

    Returns:
        torch.Tensor: The cosine angle between the two vectors.
    """
    with torch.no_grad():
        return torch.dot(vector1,vector2)/vector1.norm()/vector2.norm()
    
def unit_vector(vector: torch.Tensor, warn_zero=False)->torch.Tensor:
    """
    Compute the unit vector of a given tensor.

    Parameters:
        vector (torch.Tensor): The input tensor.
        warn_zero (bool): Whether to print a warning when the input tensor is zero. default: False

    Returns:
        torch.Tensor: The unit vector of the input tensor.
    """
    with torch.no_grad():
        if vector.norm()==0:
            if warn_zero:
                print("Detected zero vector when doing normalization.")
            return torch.zeros_like(vector)
        else:
            return vector / vector.norm()

def transfer_coef_double(weights: torch.tensor, 
                         unit_vec_1: torch.tensor, unit_vec_2: torch.tensor,
                         or_unit_vec_1: torch.tensor, or_unit_vec_2: torch.tensor)-> tuple:
    """
    Transfer the angle weights to a length coefficient for ConFIG method with two vectors.
    
    This function will return a coefficient matrix [c_1,c_2] so that
    $$
    \frac{(c_1 o_1+c_2 o_2)\dot \mathbf{v}_1}{(c_1 o_1+c_2 o_2) \dot \mathbf{v}_2}
    =\frac{w_1}{w_2}
    $$
    where w_1 and w_2 are the angle weights,
    v_1 and v_2 are unit vectors of the two gradients,
    and o_1 and o_2 are the unit orthogonal components of the two gradients.

    Args:
        weights (torch.tensor): Angle weights for the two vectors. It should have shape (2,).
        unit_vecs (torch.tensor): Unit vectors of the orthogonal components. It should have shape (2,k) where k is the length of the vector.

    Returns:
        tuple: The coefficients for the two vectors.

    Raises:
        ValueError: If the number of weights or unit vectors is not equal to 2.
    """    

    return torch.dot(or_unit_vec_2,unit_vec_1)/(weights[0]/weights[1]*torch.dot(or_unit_vec_1,unit_vec_2)),1

def estimate_conflict(gradients: torch.Tensor)->torch.Tensor:
    """
    Estimates the degree of conflict of gradients.

    Args:
        gradients (torch.Tensor): A tensor containing gradients.

    Returns:
        torch.Tensor: A tensor consistent of the dot products between the sum of gradients and each sub-gradient.
    """
    direct_sum = unit_vector(gradients.sum(dim=0))
    unit_grads = gradients / torch.norm(gradients, dim=1).view(-1, 1)
    return unit_grads @ direct_sum

def has_zero(lists:Sequence)->bool:
    """
    Check if any element in the list is zero.

    Args:
        lists (Sequence): A list of elements.

    Returns:
        bool: True if any element is zero, False otherwise.
    """
    for i in lists:
        if i==0:
            return True
    return False

class OrderedSliceSelector:
    """
    Selects a slice of the source sequence in order.
    Usually used for selecting loss functions/gradients/losses in momentum-based method if you want to update more tha one gradient in a single iteration.
    
    """
    def __init__(self):
        self.start_index=0
        
    def select(self, n:int, source_sequence:Sequence) -> Tuple[Sequence,Union[float,Sequence]]:
        """
        Selects a slice of the source sequence in order.
        
        Args:
            n (int): The length of the target slice.
            source_sequence (Sequence): The source sequence to select from.
        
        Returns:
            Tuple[Sequence,Union[float,Sequence]]: A tuple containing the indexes of the selected slice and the selected slice.
        """
        if n > len(source_sequence):
            raise ValueError("n must be less than or equal to the length of the source sequence")
        end_index = self.start_index + n
        if end_index > len(source_sequence)-1:
            new_start=end_index-len(source_sequence)
            indexes = list(range(self.start_index,len(source_sequence)))+list(range(0,new_start))
            self.start_index=new_start
        else:
            indexes = list(range(self.start_index,end_index))
            self.start_index=end_index
        if len(indexes)==1:
            return indexes,source_sequence[indexes[0]]
        else:
            return indexes,[source_sequence[i] for i in indexes]
        
class RandomSliceSelector:
    """
    Selects a slice of the source sequence randomly.
    Usually used for selecting loss functions/gradients/losses in momentum-based method if you want to update more tha one gradient in a single iteration.
    """
        
    def select(self, n:int, source_sequence:Sequence)-> Tuple[Sequence,Union[float,Sequence]]:
        """
        Selects a slice of the source sequence randomly.
        
        Args:
            n (int): The length of the target slice.
            source_sequence (Sequence): The source sequence to select from.
        
        Returns:
            Tuple[Sequence,Union[float,Sequence]]: A tuple containing the indexes of the selected slice and the selected slice.
        """
        if n > len(source_sequence):
            raise ValueError("n must be less than or equal to the length of the source sequence")
        indexes = np.random.choice(len(source_sequence),n,replace=False)
        if len(indexes)==1:
            return indexes,source_sequence[indexes[0]]
        else:
            return indexes,[source_sequence[i] for i in indexes]


from . import *
class LossRecorder:
    """
    Base class for loss recorders.
    
    Args:
        num_losses (int): The number of losses to record
    """
    
    def __init__(self,num_losses:int) -> None:
        self.num_losses=num_losses
        self.current_losses=[0.0 for i in range(num_losses)]
    
    def record_loss(self,
                    losses_indexes:Union[int,Sequence[int]], 
                    losses: Union[float,Sequence]) -> list:
            """
            Records the given loss and returns the recorded losses.

            Args:
                losses_indexes: The index of the loss.
                losses (torch.Tensor): The loss to record.

            Returns:
                list: The recorded losses.

            Raises:
                NotImplementedError: If the method is not implemented.

            """
            raise NotImplementedError("record_loss method must be implemented")
    
    def record_all_losses(self,losses: Sequence) -> list:
        """
        Records all the losses and returns the recorded losses.

        Args:
            losses (torch.Tensor): The losses to record.

        Returns:
            list: The recorded losses.

        """
        if len(losses)!=self.num_losses:
            raise ValueError("The number of losses does not match the number of losses to be recorded.")
        return self.record_loss([i for i in range(self.num_losses)],losses)
    
    def _preprocess_losses(self,
                           losses_indexes: Union[int, Sequence[int]],
                           losses: Union[float, Sequence]) -> Tuple[Sequence[int], Sequence]:
        """
        Preprocesses the losses and their indexes. Recommended to be used in the `record_loss` method.

        Args:
            losses_indexes (Union[int, Sequence[int]]): The indexes of the losses.
            losses (Union[float, Sequence]): The losses.

        Returns:
            Tuple[Sequence[int], Sequence]: A tuple containing the preprocessed losses indexes and losses.
        """
        if isinstance(losses_indexes, int):
            losses_indexes = [losses_indexes]
        if isinstance(losses, float):
            losses = [losses]
        return losses_indexes, losses
    
class LatestLossRecorder(LossRecorder):
    """
    A loss recorder return the latest losses.

    Args:
        num_losses (int): The number of losses to record
    """
    
    def __init__(self,num_losses:int) -> None:
        super().__init__(num_losses)
    
    def record_loss(self,
                    losses_indexes:Union[int,Sequence[int]], 
                    losses: Union[float,Sequence]) -> list:
        """
        Records the given loss and returns the recorded loss.

        Args:
            losses_indexes: The index of the loss.
            losses (torch.Tensor): The loss to record.

        Returns:
            list: The recorded loss.

        """
        losses_indexes,losses=self._preprocess_losses(losses_indexes,losses)
        for i in losses_indexes:
            self.current_losses[i]=losses[losses_indexes.index(i)]
        return self.current_losses
    
class MomentumLossRecorder(LossRecorder):
    """
    A loss recorder that records the momentum of the loss.
    
    Args:
        num_losses (int): The number of losses to record
        betas (Union[float, Sequence[float]]): The moving average constant.
    """
    
    def __init__(self, num_losses:int, betas: Union[float, Sequence[float]] = 0.9):
        super().__init__(num_losses)
        if isinstance(betas, float):
            self.betas = [betas] * num_losses
        self.m=[0.0 for i in range(num_losses)]
        self.t=[0 for i in range(num_losses)]
    
    def record_loss(self,
                    losses_indexes:Union[int,Sequence[int]], 
                    losses: Union[float,Sequence]) -> list:
        """
        Records the given loss and returns the recorded loss.

        Args:
            losses_indexes: The index of the loss.
            losses (torch.Tensor): The loss to record.

        Returns:
            list: The recorded loss.

        """
        losses_indexes,losses=self._preprocess_losses(losses_indexes,losses)
        for index in losses_indexes:
            self.t[index]+=1
            self.m[index]=self.betas*self.m[index]+(1-self.betas[index])*losses[losses_indexes.index(index)]
        self.current_losses=[self.m[index]/(1-self.betas[index]**self.t[index]) for index in len(self.m)]
        return self.current_losses


from torch import Tensor
from torch.nn.modules import Module
from conflictfree import Sequence
from conflictfree.grad_operator import ConFIGOperator, GradientOperator
from conflictfree.utils import Sequence
from conflictfree.loss_recorder import LatestLossRecorder, LossRecorder
from . import *
from .utils import *
from .loss_recorder import *
from .grad_operator import *

class MomentumOperator():
    """
    The base class for the momentum based operator.
    
    Args:
        num_vectors (int): The number of gradient vectors.
        beta_1 (float): The moving average constant for the first momentum.
        beta_2 (float): The moving average constant for the second momentum.
        gradient_operator (GradientOperator, optional): The base gradient operator. Defaults to ConFIGOperator().
        loss_recorder (LossRecorder, optional): The loss recorder object. 
            If you want to pass loss information to "update_gradient" method or "apply_gradient" method, you need to specify a loss recorder. Defaults to None.

    Methods:
        calculate_gradient(indexes, grads, losses=None):
            Calculates the gradient based on the given indexes, gradients, and losses.
        update_gradient(network, indexes, grads, losses=None):
            Updates the gradient of the given network with the calculated gradient.
    """

    def __init__(self, 
                 num_vectors: int, 
                 beta_1: float=0.9, 
                 beta_2: float=0.999,
                 gradient_operator: GradientOperator=ConFIGOperator(),
                 loss_recorder: Optional[LossRecorder] = None) -> None:
        self.len_vectors = None
        self.device = None
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.num_vectors = num_vectors
        self.gradient_operator = gradient_operator
        self.loss_recorder = loss_recorder

    def calculate_gradient(self, 
                           indexes: Union[int,Sequence[int]], 
                           grads: Union[torch.Tensor,Sequence[torch.Tensor]], 
                           losses: Optional[Union[float,Sequence]] = None) -> torch.Tensor:
        """
        Calculates the gradient based on the given indexes, gradients, and losses.

        Args:
            indexes (Union[int,Sequence[int]]): The indexes of the gradient vectors and losses to be updated.
                The momentum with the given indexes will be updated based on the given gradients.
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to base gradient operator. If the base gradient operator doesn't require loss information,
                you can set this value as None. Defaults to None.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
            
        Returns:
            torch.Tensor: The calculated gradient.
        """

        raise NotImplementedError("calculate_gradient method must be implemented")

    def _preprocess_gradients_losses(self, 
                                     indexes: Union[int,Sequence[int]], 
                                     grads: Union[torch.Tensor,Sequence[torch.Tensor]], 
                                     losses: Optional[Union[float,Sequence]] = None) -> tuple:
        """
        Preprocesses the gradients and losses before applying momentum.
        Recommended to be used in the `update_gradient` method.
        
        Args:
            indexes (Union[int,Sequence[int]]): The indexes of the gradient vectors and losses to be updated.
                The momentum with the given indexes will be updated based on the given gradients.
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to base gradient operator. If the base gradient operator doesn't require loss information,
                you can set this value as None. Defaults to None.

        Returns:
            Tuple: A tuple containing the preprocessed indexes, gradients, and losses (if provided).
        """
        if not isinstance(indexes,Sequence):
            indexes=[indexes]
        if isinstance(grads,torch.Tensor):
            grads=[grads]
        if losses is not None:
            if self.loss_recorder is None:
                raise ValueError("Losses provided but loss_recorder is not provided.")
            if not isinstance(losses,Sequence):
                losses=[losses]
            losses=self.loss_recorder.record_loss(indexes,losses)
        if self.len_vectors is None or self.device is None:
            self.len_vectors = grads[0].shape[0]
            self.device = grads[0].device
        return indexes,grads,losses
        

    def update_gradient(self, network: torch.nn.Module, 
                        indexes: Union[int,Sequence[int]], 
                        grads: Union[torch.Tensor,Sequence[torch.Tensor]],
                        losses: Optional[Union[float,Sequence]] = None) -> None:
        """
        Updates the gradient of the given network with the calculated gradient.

        Args:
            network (torch.nn.Module): The network to update the gradient.
            indexes (Union[int,Sequence[int]]): The indexes of the gradient vectors and losses to be updated.
                The momentum with the given indexes will be updated based on the given gradients.
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to base gradient operator. If the base gradient operator doesn't require loss information,
                you can set this value as None. Defaults to None.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
            
        Returns:
            None
        """
        apply_gradient_vector(network, self.calculate_gradient(indexes, grads, losses))


class PseudoMomentumOperator(MomentumOperator):

    """
    The major momentum version. 
    In this operator, the second momentum is estimated by a pseudo gradient based on the result of the gradient operator.
    NOTE: The momentum-based operator, e.g., Adam, is not recommend when using this operator. Please consider using SGD optimizer.
    
    Args:
        num_vectors (int): The number of gradient vectors.
        beta_1 (float): The moving average constant for the first momentum.
        beta_2 (float): The moving average constant for the second momentum.
        gradient_operator (GradientOperator, optional): The base gradient operator. Defaults to ConFIGOperator().
        loss_recorder (LossRecorder, optional): The loss recorder object. 
            If you want to pass loss information to "update_gradient" method or "apply_gradient" method, you need to specify a loss recorder. Defaults to None.

    Methods:
        calculate_gradient(indexes, grads, losses=None):
            Calculates the gradient based on the given indexes, gradients, and losses.
        update_gradient(network, indexes, grads, losses=None):
            Updates the gradient of the given network with the calculated gradient.
            
    Examples:
    ```python
    from conflictfree.momentum_operator import PseudoMomentumOperator
    from conflictfree.utils import get_gradient_vector,apply_gradient_vector
    optimizer=torch.Adam(network.parameters(),lr=1e-3)
    operator=PseudoMomentumOperator(num_vector=len(loss_fns)) # initialize operator, the only difference here is we need to specify the number of gradient vectors.
    for input_i in dataset:
        grads=[]
        for loss_fn in loss_fns:
            optimizer.zero_grad()
            loss_i=loss_fn(input_i)
            loss_i.backward()
            grads.append(get_gradient_vector(network))
        g_config=operator.calculate_gradient(grads) # calculate the conflict-free direction
        apply_gradient_vector(network) # or simply use `operator.update_gradient(network,grads)` to calculate and set the condlict-free direction to the network
        optimizer.step()
    ```
    """
    
    def __init__(self, 
                num_vectors: int, 
                 beta_1: float=0.9, 
                 beta_2: float=0.999,
                 gradient_operator: GradientOperator=ConFIGOperator(),
                 loss_recorder: Optional[LossRecorder] = None) -> None:
        super().__init__(num_vectors, beta_1, beta_2, gradient_operator, loss_recorder)
        self.m=None
        self.s=None
        self.fake_m=None
        self.t=0
        self.t_grads=[0]*self.num_vectors
        self.all_initialized=False        
    
    def _preprocess_gradients_losses(self,
                                     indexes: Union[int,Sequence[int]], 
                                     grads: Union[torch.Tensor,Sequence[torch.Tensor]], 
                                     losses: Optional[Union[float,Sequence]] = None):
        indexes,grads,losses=super()._preprocess_gradients_losses(indexes, grads, losses)
        if self.m is None or self.s is None or self.fake_m is None:
            self.m=[torch.zeros(self.len_vectors,device=self.device) for i in range(self.num_vectors)]
            self.s=torch.zeros(self.len_vectors,device=self.device)
            self.fake_m=torch.zeros(self.len_vectors,device=self.device)        
        return indexes,grads,losses
    
    def calculate_gradient(self, 
                           indexes: Union[int,Sequence[int]], 
                           grads: Union[torch.Tensor,Sequence[torch.Tensor]], 
                           losses: Optional[Union[float,Sequence]] = None)->torch.Tensor:
        """
        Calculates the gradient based on the given indexes, gradients, and losses.

        Args:
            indexes (Union[int,Sequence[int]]): The indexes of the gradient vectors and losses to be updated.
                The momentum with the given indexes will be updated based on the given gradients.
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to base gradient operator. If the base gradient operator doesn't require loss information,
                you can set this value as None. Defaults to None.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
            
        Returns:
            torch.Tensor: The calculated gradient.
        """
        with torch.no_grad():
            indexes,grads,losses=self._preprocess_gradients_losses(indexes,grads,losses)
            for i in range(len(indexes)):
                self.t_grads[indexes[i]]+=1
                self.m[indexes[i]]=self.beta_1*self.m[indexes[i]]+(1-self.beta_1)*grads[i]
            if not self.all_initialized:
                if has_zero(self.t_grads):
                    return torch.zeros_like(self.s)
                else:
                    self.all_initialized=True
            self.t+=1
            m_hats=torch.stack([self.m[i]/(1-self.beta_1**self.t_grads[i]) for i in range(self.num_vectors)],dim=0)
            final_grad=self.gradient_operator.calculate_gradient(m_hats,
                                                                 losses
                                                                 )
            fake_m=final_grad*(1-self.beta_1**self.t)
            fake_grad=(fake_m-self.beta_1*self.fake_m)/(1-self.beta_1)
            self.fake_m=fake_m
            self.s=self.beta_2*self.s+(1-self.beta_2)*fake_grad**2
            s_hat=self.s/(1-self.beta_2**self.t)
            final_grad=final_grad/(torch.sqrt(s_hat)+1e-8)
        return final_grad
    
class SeparateMomentumOperator(MomentumOperator):
    
    """
    In this operator, each gradient has its own second gradient. The gradient operator is applied on the rescaled momentum.
    NOTE: Please consider using the PseudoMomentumOperator since this operator does not give good performance according to our research.
    The momentum-based operator, e.g., Adam, is not recommend when using this operator. Please consider using SGD optimizer.
    
    Args:
        num_vectors (int): The number of gradient vectors.
        beta_1 (float): The moving average constant for the first momentum.
        beta_2 (float): The moving average constant for the second momentum.
        gradient_operator (GradientOperator, optional): The base gradient operator. Defaults to ConFIGOperator().
        loss_recorder (LossRecorder, optional): The loss recorder object. 
            If you want to pass loss information to "update_gradient" method or "apply_gradient" method, you need to specify a loss recorder. Defaults to None.

    Methods:
        calculate_gradient(indexes, grads, losses=None):
            Calculates the gradient based on the given indexes, gradients, and losses.
        update_gradient(network, indexes, grads, losses=None):
            Updates the gradient of the given network with the calculated gradient.
            
    """
    
    def __init__(self, 
                num_vectors: int, 
                 beta_1: float=0.9, 
                 beta_2: float=0.999,
                 gradient_operator: GradientOperator=ConFIGOperator(),
                 loss_recorder: Optional[LossRecorder] = None) -> None:
        super().__init__(num_vectors, beta_1, beta_2, gradient_operator, loss_recorder)
        self.m=None
        self.s=None
        self.t_grads=[0]*len(self.num_vectors)
        self.all_initialized=False        
        
    def _preprocess_gradients_losses(self,
                                     indexes: Union[int,Sequence[int]], 
                                     grads: Union[torch.Tensor,Sequence[torch.Tensor]], 
                                     losses: Optional[Union[float,Sequence]] = None):
        indexes,grads,losses=super()._preprocess_gradients_losses(indexes, grads, losses)
        if self.m is None or self.s is None:
            self.m=[torch.zeros(self.len_vectors,device=self.device) for i in range(self.num_vectors)]
            self.s=[torch.zeros(self.len_vectors,device=self.device) for i in range(self.num_vectors)]   
        return indexes,grads,losses
    
    def calculate_gradient(self, 
                           indexes: Union[int,Sequence[int]], 
                           grads: Union[torch.Tensor,Sequence[torch.Tensor]], 
                           losses: Optional[Union[float,Sequence]] = None)->torch.Tensor:
        """
        Calculates the gradient based on the given indexes, gradients, and losses.

        Args:
            indexes (Union[int,Sequence[int]]): The indexes of the gradient vectors and losses to be updated.
                The momentum with the given indexes will be updated based on the given gradients.
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to base gradient operator. If the base gradient operator doesn't require loss information,
                you can set this value as None. Defaults to None.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
            
        Returns:
            torch.Tensor: The calculated gradient.
        """
        with torch.no_grad():
            indexes,grads,losses=self._preprocess_gradients_losses(indexes,grads,losses)
            for i in range(len(indexes)):
                self.t_grads[indexes[i]]+=1
                self.m[indexes[i]]=self.beta_1*self.m[indexes[i]]+(1-self.beta_1)*grads[i]
                self.s[indexes[i]]=self.beta_2*self.s[indexes[i]]+(1-self.beta_2)*grads[i]**2
            if not self.all_initialized:
                if has_zero(self.t_grads):
                    return torch.zeros_like(self.s)
                else:
                    self.all_initialized=True
            m_hats=torch.stack([self.m[i]/(1-self.betas_1**self.t_grads[i]) for i in range(self.num_vectors)],dim=0)
            s_hats=torch.stack([self.s[i]/(1-self.betas_2**self.t_grads[i]) for i in range(self.num_vectors)],dim=0)
        return self.gradient_operator.calculate_gradient(m_hats/(torch.sqrt(s_hats)+1e-8),
                                                         losses,)


import setuptools
from pathlib import Path

with open("pypi_desc.md", "r") as fh:
    long_description = fh.read()

def get_install_requires():
    """Returns requirements.txt parsed to a list"""
    fname = Path(__file__).parent / 'requirements.txt'
    targets = []
    if fname.exists():
        with open(fname, 'r') as f:
            targets = f.read().splitlines()
    return targets

setuptools.setup(
    name="conflictfree",
    version="0.1.4",
    author="Qiang Liu, Mengyu Chu, Nils Thuerey",
    author_email="qiangliu.7@outlook.com",
    description="Official implementation of Conflict-free Inverse Gradients method",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://tum-pbs.github.io/ConFIG/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=get_install_requires(),
)



from . import *
from .utils import *


class LengthModel:
    """
    The base class for length model.

    Methods:
        get_length: Calculates the length based on the given parameters.
        rescale_length: Rescales the length of the target vector based on the given parameters.
    """
    def __init__(self):
        pass
    
    def get_length(self, 
                   target_vector:Optional[torch.Tensor]=None,
                   unit_target_vector:Optional[torch.Tensor]=None,
                   gradients:Optional[torch.Tensor]=None,
                   losses:Optional[Sequence]=None)-> Union[torch.Tensor, float]:
        """
        Calculates the length based on the given parameters. Not all parameters are required.

        Args:
            target_vector (Optional[torch.Tensor]): The final update gradient vector.
            unit_target_vector (Optional[torch.Tensor]): The unit vector of the target vector.
            gradients (Optional[torch.Tensor]): The loss-specific gradients matrix.
            losses (Optional[Sequence]): The losses.

        Returns:
            Union[torch.Tensor, float]: The calculated length.
        """
        raise NotImplementedError("This method must be implemented by the subclass.")
    
    def rescale_length(self, 
                       target_vector:torch.Tensor,
                       gradients:Optional[torch.Tensor]=None,
                       losses:Optional[Sequence]=None)->torch.Tensor:
        """
        Rescales the length of the target vector based on the given parameters.
        It calls the get_length method to calculate the length and then rescales the target vector.
        
        Args:
            target_vector (torch.Tensor): The final update gradient vector.
            gradients (Optional[torch.Tensor]): The loss-specific gradients matrix.
            losses (Optional[Sequence]): The losses.
        
        Returns:
            torch.Tensor: The rescaled target vector.
        """
        unit_target_vector = unit_vector(target_vector)
        return self.get_length(target_vector=target_vector,
                               unit_target_vector=unit_target_vector,
                               gradients=gradients,
                               losses=losses) * unit_target_vector
    
class ProjectionLength(LengthModel):
    """
    Rescale the length of the target vector based on the projection of the gradients on the target vector.
    """

    def __init__(self):
        super().__init__()
        
    def get_length(self, target_vector:Optional[torch.Tensor]=None,
                         unit_target_vector:Optional[torch.Tensor]=None,
                         gradients:Optional[torch.Tensor]=None,
                         losses:Optional[Sequence]=None)->torch.Tensor:
        """
        Calculates the length based on the given parameters. Not all parameters are required.

        Args:
            target_vector (Optional[torch.Tensor]): The final update gradient vector. 
                One of the `target_vector` or `unit_target_vector` parameter need to be provided.
            unit_target_vector (Optional[torch.Tensor]): The unit vector of the target vector.
                One of the `target_vector` or `unit_target_vector` parameter need to be provided.
            gradients (Optional[torch.Tensor]): The loss-specific gradients matrix.
            losses (Optional[Sequence]): The losses. Not used in this model.

        Returns:
            Union[torch.Tensor, float]: The calculated length.
        """
        if gradients is None:
            raise ValueError("The ProjectLength model requires gradients information.")
        if unit_target_vector is None:
            unit_target_vector = unit_vector(target_vector)
        return torch.sum(torch.stack([torch.dot(grad_i,unit_target_vector) for grad_i in gradients]))


from . import *

class WeightModel:
    """
    Base class for weight models. 
    """
    def __init__(self):
        pass
    
    def get_weights(self, gradients:Optional[torch.Tensor]=None,losses:Optional[Sequence]=None):
        """_summary_

        Args:
            gradients (Optional[torch.Tensor]): The loss-specific gradients matrix.
            losses (Optional[Sequence]): The losses.

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError("This method must be implemented by the subclass.")

class EqualWeight(WeightModel):
    """
    A weight model that assigns equal weights to all gradients.
    """

    def __init__(self):
        super().__init__()
    
    def get_weights(self, gradients: torch.Tensor, losses: Optional[Sequence] = None,device:Optional[Union[torch.device,str]]=None) -> torch.Tensor:
        """
        Calculate the weights for the given gradients.

        Args:
            gradients (Optional[torch.Tensor]): The loss-specific gradients matrix.
            losses (Optional[Sequence]): The losses. Not used in this model.

        Returns:
            torch.Tensor: A tensor of equal weights for all gradients.
        
        Raises:
            ValueError: If gradients is None.
        """
        if gradients is None:
            raise ValueError("The EqualWeight model requires gradients.")
        return torch.ones(gradients.shape[0],device=device)


#usr/bin/python3
# -*- coding: UTF-8 -*-
from . import *

from typing import Union
from .utils import *
from .weight_model import *
from .length_model import *

def ConFIG_update_double(grad_1:torch.Tensor,grad_2:torch.Tensor,
                         weight_model:WeightModel=EqualWeight(),
                         length_model:LengthModel=ProjectionLength(),
                         losses:Optional[Sequence]=None)-> torch.Tensor:
    """
    ConFIG update for two gradients where no inverse calculation is needed. 

    Args:
        grad_1 (torch.Tensor): The first gradient.
        grad_2 (torch.Tensor): The second gradient.
        weight_model (WeightModel, optional): The weight model for calculating the direction weights.
            Defaults to EqualWeight(), which will make the final update gradient not biased towards any gradient.
        length_model (LengthModel, optional): The length model for rescaling the length of the final gradient. 
            Defaults to ProjectionLength(), which will project each gradient vector onto the final gradient vector to get the final length.
        losses (Optional[Sequence], optional): The losses associated with the gradients. 
            The losses will be passed to the weight and length model. If your weight/length model doesn't require loss information,
            you can set this value as None. Defaults to None.

    Returns:
        torch.Tensor: The final update gradient.
        
    Examples:
        ```python
        from conflictfree.grad_operator import ConFIG_update_double
        from conflictfree.utils import get_gradient_vector,apply_gradient_vector
        optimizer=torch.Adam(network.parameters(),lr=1e-3)
        for input_i in dataset:
            grads=[] # we record gradients rather than losses
            for loss_fn in [loss_fn1, loss_fn2]:
                optimizer.zero_grad()
                loss_i=loss_fn(input_i)
                loss_i.backward()
                grads.append(get_gradient_vector(network)) #get loss-specfic gradient
            g_config=ConFIG_update_double(grads) # calculate the conflict-free direction
            apply_gradient_vector(network) # set the condlict-free direction to the network
            optimizer.step()
        ```

    """
    with torch.no_grad():
        norm_1=grad_1.norm();norm_2=grad_2.norm()
        unit_1=grad_1/norm_1;unit_2=grad_2/norm_2
        cos_angle=get_cos_similarity(grad_1,grad_2)
        or_2=grad_1-norm_1*cos_angle*unit_2
        or_1=grad_2-norm_2*cos_angle*unit_1
        unit_or1=unit_vector(or_1);unit_or2=unit_vector(or_2)
        coef_1,coef_2=transfer_coef_double(
            weight_model.get_weights(gradients=torch.stack([grad_1,grad_2]),losses=losses,device=grad_1.device),
            unit_1,
            unit_2,
            unit_or1,
            unit_or2,)
        best_direction=coef_1*unit_or1+coef_2*unit_or2
        return length_model.rescale_length(target_vector=best_direction,
                                           gradients=torch.stack([grad_1,grad_2]),
                                           losses=losses,
                                           )

def ConFIG_update(
    grads:Union[torch.Tensor,Sequence[torch.Tensor]],
    weight_model:WeightModel=EqualWeight(),
    length_model:LengthModel=ProjectionLength(),
    use_least_square:bool=True,
    losses:Optional[Sequence]=None)-> torch.Tensor:
    """
    Performs the standard ConFIG update step.

    Args:
        grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
            It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
        weight_model (WeightModel, optional): The weight model for calculating the direction weights.
            Defaults to EqualWeight(), which will make the final update gradient not biased towards any gradient.
        length_model (LengthModel, optional): The length model for rescaling the length of the final gradient. 
            Defaults to ProjectionLength(), which will project each gradient vector onto the final gradient vector to get the final length.
        use_least_square (bool, optional): Whether to use the least square method for calculating the best direction. 
            If set to False, we will directly calculate the pseudo-inverse of the gradient matrix. See `torch.linalg.pinv` and `torch.linalg.lstsq` for more details.
            Recommended to set to True. Defaults to True.
        losses (Optional[Sequence], optional): The losses associated with the gradients. 
            The losses will be passed to the weight and length model. If your weight/length model doesn't require loss information,
            you can set this value as None. Defaults to None.

    Returns:
        torch.Tensor: The final update gradient.
        
    Examples:
        ```python
        from conflictfree.grad_operator import ConFIG_update
        from conflictfree.utils import get_gradient_vector,apply_gradient_vector
        optimizer=torch.Adam(network.parameters(),lr=1e-3)
        for input_i in dataset:
            grads=[] # we record gradients rather than losses
            for loss_fn in loss_fns:
                optimizer.zero_grad()
                loss_i=loss_fn(input_i)
                loss_i.backward()
                grads.append(get_gradient_vector(network)) #get loss-specfic gradient
            g_config=ConFIG_update(grads) # calculate the conflict-free direction
            apply_gradient_vector(network) # set the condlict-free direction to the network
            optimizer.step()
        ```
    """
    if not isinstance(grads,torch.Tensor):
        grads=torch.stack(grads)
    with torch.no_grad():
        weights=weight_model.get_weights(gradients=grads,losses=losses,device=grads.device)
        units=torch.nan_to_num((grads/(grads.norm(dim=1)).unsqueeze(1)),0)
        if use_least_square:
            best_direction=torch.linalg.lstsq(units, weights).solution
        else:
            best_direction=torch.linalg.pinv(units)@weights
        return length_model.rescale_length(target_vector=best_direction,
                                           gradients=grads,
                                           losses=losses,
                                            )

class GradientOperator:
    """
    A base class that represents a gradient operator.

    Methods:
        calculate_gradient: Calculates the gradient based on the given gradients and losses.
        update_gradient: Updates the gradient of the network based on the calculated gradient.

    """

    def __init__(self):
        pass
    
    def calculate_gradient(self, grads: Union[torch.Tensor,Sequence[torch.Tensor]], losses: Optional[Sequence] = None)-> torch.Tensor:
        """
        Calculates the gradient based on the given gradients and losses.

        Args:
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to the weight and length model. If your weight/length model doesn't require loss information,
                you can set this value as None. Defaults to None.

        Returns:
            torch.Tensor: The calculated gradient.

        Raises:
            NotImplementedError: If the method is not implemented.

        """
        raise NotImplementedError("calculate_gradient method must be implemented")
    
    def update_gradient(self, network: torch.nn.Module, grads: Union[torch.Tensor,Sequence[torch.Tensor]], losses: Optional[Sequence] = None)-> None:
        """
        Calculate the gradient and apply the gradient to the network.

        Args:
            network (torch.nn.Module): The target network.
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to the weight and length model. If your weight/length model doesn't require loss information,
                you can set this value as None. Defaults to None.

        Returns:
            None

        """
        apply_gradient_vector(network, self.calculate_gradient(grads, losses))
    

class ConFIGOperator(GradientOperator):
    """
    Operator for the ConFIG algorithm.

    Args:
        weight_model (WeightModel, optional): The weight model for calculating the direction weights.
            Defaults to EqualWeight(), which will make the final update gradient not biased towards any gradient.
        length_model (LengthModel, optional): The length model for rescaling the length of the final gradient. 
            Defaults to ProjectionLength(), which will project each gradient vector onto the final gradient vector to get the final length.
        allow_simplified_model (bool, optional): Whether to allow simplified model for calculating the gradient. 
            If set to True, will use simplified form of ConFIG method when there are only two losses (ConFIG_update_double). Defaults to True.
        use_least_square (bool, optional): Whether to use the least square method for calculating the best direction. 
            If set to False, we will directly calculate the pseudo-inverse of the gradient matrix. See `torch.linalg.pinv` and `torch.linalg.lstsq` for more details.
            Recommended to set to True. Defaults to True.
    
    Examples:
        ```python
        from conflictfree.grad_operator import ConFIGOperator
        from conflictfree.utils import get_gradient_vector,apply_gradient_vector
        optimizer=torch.Adam(network.parameters(),lr=1e-3)
        operator=ConFIGOperator() # initialize operator
        for input_i in dataset:
            grads=[]
            for loss_fn in loss_fns:
                optimizer.zero_grad()
                loss_i=loss_fn(input_i)
                loss_i.backward()
                grads.append(get_gradient_vector(network))
            g_config=operator.calculate_gradient(grads) # calculate the conflict-free direction
            apply_gradient_vector(network) # or simply use `operator.update_gradient(network,grads)` to calculate and set the condlict-free direction to the network
            optimizer.step()
        ```
    
    """

    def __init__(self, 
                 weight_model: WeightModel = EqualWeight(),
                 length_model: LengthModel = ProjectionLength(),
                 allow_simplified_model: bool = True,
                 use_least_square: bool = True):
        super().__init__()
        self.weight_model = weight_model
        self.length_model = length_model
        self.allow_simplified_model = allow_simplified_model
        self.use_least_square = use_least_square

    def calculate_gradient(self, grads: Union[torch.Tensor,Sequence[torch.Tensor]], losses: Optional[Sequence] = None)->torch.Tensor:
        """
        Calculates the gradient using the ConFIG algorithm.

        Args:
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): The losses associated with the gradients. 
                The losses will be passed to the weight and length model. If your weight/length model doesn't require loss information,
                you can set this value as None. Defaults to None.

        Returns:
            torch.Tensor: The calculated gradient.
        """
        if not isinstance(grads,torch.Tensor):
            grads=torch.stack(grads)
        if grads.shape[0] == 2 and self.allow_simplified_model:
            return ConFIG_update_double(grads[0], grads[1],
                                        weight_model=self.weight_model,
                                        length_model=self.length_model,
                                        losses=losses)
        else:
            return ConFIG_update(grads,
                                 weight_model=self.weight_model,
                                 length_model=self.length_model,
                                 use_least_square=self.use_least_square,
                                 losses=losses)

class PCGradOperator(GradientOperator):
    """
    PCGradOperator class represents a gradient operator for PCGrad algorithm.
    
    @inproceedings{yu2020gradient,
    title={Gradient surgery for multi-task learning},
    author={Yu, Tianhe and Kumar, Saurabh and Gupta, Abhishek and Levine, Sergey and Hausman, Karol and Finn, Chelsea},
    booktitle={34th International Conference on Neural Information Processing Systems},
    year={2020},
    url={https://dl.acm.org/doi/abs/10.5555/3495724.3496213}
    }

    """


    def calculate_gradient(self, grads: Union[torch.Tensor,Sequence[torch.Tensor]], losses: Optional[Sequence] = None)->torch.Tensor:
        """
        Calculates the gradient using the PCGrad algorithm.

        Args:
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): This parameter should not be set for current operator. Defaults to None.

        Returns:
            torch.Tensor: The calculated gradient using PCGrad method.
        """
        if not isinstance(grads,torch.Tensor):
            grads=torch.stack(grads)
        with torch.no_grad():
            grads_pc=torch.clone(grads)
            length=grads.shape[0]
            for i in range(length):
                for j in range(length):
                    if j !=i:
                        dot=grads_pc[i].dot(grads[j])
                        if dot<0:
                            grads_pc[i]-=dot*grads[j]/((grads[j].norm())**2)
            return torch.sum(grads_pc,dim=0)
        
class IMTLGOperator(GradientOperator):
    """
    PCGradOperator class represents a gradient operator for IMTL-G algorithm.
    
    @inproceedings{
    liu2021towards,
    title={Towards Impartial Multi-task Learning},
    author={Liyang Liu and Yi Li and Zhanghui Kuang and Jing-Hao Xue and Yimin Chen and Wenming Yang and Qingmin Liao and Wayne Zhang},
    booktitle={International Conference on Learning Representations},
    year={2021},
    url={https://openreview.net/forum?id=IMPnRXEWpvr}
    }

    """


    def calculate_gradient(self, grads: Union[torch.Tensor,Sequence[torch.Tensor]], losses: Optional[Sequence] = None) ->torch.Tensor:
        """
        Calculates the gradient using the IMTL-G algorithm.

        Args:
            grads (Union[torch.Tensor,Sequence[torch.Tensor]]): The gradients to update. 
                It can be a stack of gradient vectors (at dim 0) or a sequence of gradient vectors.
            losses (Optional[Sequence], optional): This parameter should not be set for current operator. Defaults to None.

        Returns:
            torch.Tensor: The calculated gradient using IMTL-G method.
        """
        if not isinstance(grads,torch.Tensor):
            grads=torch.stack(grads)
        with torch.no_grad():
            ut_norm=grads/grads.norm(dim=1).unsqueeze(1)
            ut_norm=torch.nan_to_num(ut_norm,0)
            ut=torch.stack([ut_norm[0]-ut_norm[i+1] for i in range(grads.shape[0]-1)],dim=0).T
            d=torch.stack([grads[0]-grads[i+1] for i in range(grads.shape[0]-1)],dim=0)
            at=grads[0]@ut@torch.linalg.pinv(d@ut)
            return (1-torch.sum(at))*grads[0]+torch.sum(at.unsqueeze(1)*grads[1:],dim=0)